"""
switch("foo").case("bar").then(
    ...
).case("baz").then(
    ...
).otherwise(
    ...
)

"""
from microcosm_pubsub.chain import Chain
from microcosm_pubsub.chain.statements.case import CaseStatement


class SwitchStatement:
    """
    Switch on one or more cases.

    """
    def __init__(self, key):
        self.key = key
        self._otherwise = None
        self._cases = dict()

    def __str__(self):
        return f"switch_{self.key}"

    def case(self, key, *args, **kwargs):
        if not args and not kwargs:
            return CaseStatement(self, key)

        self._cases[key] = Chain.make(*args, **kwargs)
        return self

    def otherwise(self, *args, **kwargs):
        self._otherwise = Chain.make(*args, **kwargs)
        return self

    def __call__(self, context):
        key = context[self.key]
        action = self._cases.get(key, self._otherwise)
        if action:
            return action(context)


def switch(key):
    """
    Run one of number of chains chains - based on a key value
    Example: when("arg")
        .case(100, Chain(...))
        .case(200, Chain(...))
        .otherwise(Chain(...))

    """
    return SwitchStatement(key)



"""
Message context.

"""


def sqs_message_context(message_dct, **kwargs):
    context = message_dct.get("opaque_data", dict())

    # If there is a uri add it to the context
    if message_dct.get("uri"):
        context["uri"] = message_dct.get("uri")

    context.update(**kwargs)
    return context


def configure_sqs_message_context(graph):
    """
    Configure the message context function which controls what data you want to associate
    with your daemon handler context, e.g. some combination of keys from the message or
    additional metadata fetched from elsewhere.

    Usage:
        graph.message_context()

    """
    return sqs_message_context



from marshmallow import fields

from microcosm_pubsub.codecs import PubSubMessageSchema


class URIMessageSchema(PubSubMessageSchema):
    """
    Define a baseline message schema that points to (the URI of) a resource that experienced a lifecycle change.

    By convention, pubsub messages should be *references* to something that happened within some other
    source-of-truth (e.g. a CRUD microservice). Because there are inherent race conditions between publishing
    a message and commit a persistent transaction, message consumers are expected to call back to source-of-truth
    (e.g. via HTTP) and fetch the current resource value. In the event that message has not been committed yet,
    the consumer can retry a few times before giving up (e.g. via SQS dead-lettering).

    """
    def __init__(self, media_type, **kwargs):
        super().__init__(**kwargs)
        self.MEDIA_TYPE = media_type

    uri = fields.String(required=True)

    def deserialize_media_type(self, obj):
        return self.MEDIA_TYPE


class IdentityMessageSchema(PubSubMessageSchema):
    """
    Define a baseline message schema that points to (the identity of) a resource that experienced a lifecycle change.

    In general, these conventions prefer URI-based messages, but for some lifecycle changes (e.g. deletion),
    a URI may not be available.

    """
    def __init__(self, media_type, **kwargs):
        super().__init__(**kwargs)
        self.MEDIA_TYPE = media_type

    id = fields.String(required=True)

    def deserialize_media_type(self, obj):
        return self.MEDIA_TYPE



"""
try(
    ...
).catch("foo").then(
    ...
).catch("bar").then(
    ...
).otherwise(
    ...
)

"""
from microcosm_pubsub.chain import Chain
from microcosm_pubsub.chain.statements.case import CaseStatement


class TryChainStatement:
    def __init__(self, *args, **kwargs):
        self.chain = Chain.make(*args, **kwargs)
        self._cases = dict()
        self._otherwise = None

    def catch(self, key, *args,  **kwargs):
        if not args and not kwargs:
            return CaseStatement(self, key)

        self._cases[key] = Chain.make(*args, **kwargs)
        return self

    def otherwise(self, *args, **kwargs):
        self._otherwise = Chain.make(*args, **kwargs)
        return self

    def __call__(self, context):
        try:
            res = self.chain(context)
        except Exception as error:
            handle = self._cases.get(type(error))
            if not handle:
                raise error
            return handle(context)
        else:
            if self._otherwise:
                return self._otherwise(context)
            return res


def try_chain(*args, **kwargs):
    """
    Run one of two chains - based on a condition

    Example: try_chain(Chain(...))
        .catch(ValueError, Chain(...))
        .catch(KeyError, Chain(...))
        .otherwise(Chain(...))

    """
    return TryChainStatement(*args, **kwargs)



"""
assign("foo.bar").to("baz")

"""
from inspect import getfullargspec


class Reference:

    def __init__(self, name):
        self.parts = name.split(".")

    @property
    def key(self):
        return self.parts[0]

    def __call__(self, context):
        value = context[self.key]

        for part in self.parts[1:]:
            if hasattr(value, part):
                value = getattr(value, part)
            else:
                value = value[part]

        return value


class Constant:

    def __init__(self, value):
        self.value = value

    def __call__(self, context):
        return self.value


class Function:

    def __init__(self, func):
        self.func = func
        try:
            self.args = getfullargspec(func).args
        except TypeError:
            # NB: getfullargspec fails for builtins like 'dict'
            #
            # And oddly `inspect.isbuiltin` doesn't work as expected either. We could instead
            # check `func.__module__ == 'builtins'` but that feels more fragile than just assuming
            # an error indictates a builtin... and that a builtin won't take our 'context' argument
            # in any useful way.
            #
            # See: https://bugs.python.org/issue1748064
            self.args = ()

    def __call__(self, context):
        if self.args:
            return self.func(context)
        else:
            return self.func()


class AssignStatement:
    """
    Assign `this` value as `that`.

    """
    def __init__(self, this, that=None):
        self.this = this
        self.that = that

    def to(self, that):
        self.that = that
        return self

    @property
    def name(self):
        return self.that

    def __str__(self):
        return f"assign_{self.name}"

    def __call__(self, context):
        value = self.this(context)
        context[self.name] = value
        return value


def assign(this):
    return AssignStatement(Reference(this))


def assign_constant(this):
    return AssignStatement(Constant(this))


def assign_function(this):
    return AssignStatement(Function(this))


def extract(name, key, key_property=None):
    """
    Extract an argument from a context to anotherwise context key

    :param name: new context key
    :param key: old context key
    :param key_property: propery of the context key

    """
    if key_property:
        key = ".".join([key, key_property])

    return AssignStatement(Reference(key), name)



"""
Message encoding and decoding.

"""
from json import dumps, loads

from marshmallow import fields, Schema, ValidationError


DEFAULT_MEDIA_TYPE = "application/json"


class PubSubMessageSchema(Schema):
    """
    Base schema for messages, including a media type.

    """
    media_type = fields.Method(
        # called by dump
        serialize="serialize_media_type",
        # called by load
        deserialize="deserialize_media_type",
        attribute="mediaType",
        # need to set missing to non-None or marshmallow won't call the deserialize function
        missing=DEFAULT_MEDIA_TYPE,
    )
    opaque_data = fields.Dict(
        attribute="opaqueData",
        required=False,
    )

    def serialize_media_type(self, message):
        """
        Fetch the media type from the message.

        """
        try:
            return message["mediaType"]
        except KeyError:
            raise ValidationError("Message did not define a media type")

    def deserialize_media_type(self, obj):
        """
        Return a custom media type.

        Should be overridden in subclasses.

        """
        return DEFAULT_MEDIA_TYPE


class MediaTypeSchema(Schema):
    """
    Custom schema for extracting media type.

    """
    mediaType = fields.String(required=True)


class PubSubMessageCodec:
    """
    Message encoder/decoder.

    """

    def __init__(self, schema):
        self.schema = schema

    def encode(self, dct=None, **kwargs):
        """
        Encode a message.

        Uses the appropriate codec to write JSON.

        """
        message = dct.copy() if dct else dict()
        message.update(kwargs)
        encoded = self.schema.load(message)
        return dumps(encoded.data)

    def decode(self, message):
        """
        Decode a message.

        Uses the appropriate coded to read JSON.

        """
        dct = loads(message)
        # we need to explicitly validate because dump() doesn't
        self.schema.validate(dct)
        decoded = self.schema.dump(dct)
        return decoded.data



from microcosm_pubsub.chain.chain import Chain  # noqa: F401
from microcosm_pubsub.chain.decorators import binds, extracts  # noqa: F401
from microcosm_pubsub.chain.statements import (  # noqa: F401
    assign,
    assign_constant,
    assign_function,
    extract,
    for_each,
    when,
    switch,
    try_chain,
)



from microcosm_pubsub.handlers.chain_handlers import ChainHandler, ChainURIHandler  # noqa: F401
from microcosm_pubsub.handlers.publish_batch_message import PublishBatchMessage  # noqa: F401
from microcosm_pubsub.handlers.uri_handler import URIHandler  # noqa: F401
